
# This file is part of FatxFast.
#
#    FatxFast is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    FatxFast is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with FatxFast.  If not, see <http://www.gnu.org/licenses/>.

# Versioning based on: 
# http://en.wikipedia.org/wiki/Versioning#Designating_development_stage
__author__ = "dryatu (c) 2013"
__version__ = "1.1.0"

import pygame
from FatxFast.tilemap.maze.mazegen import *
import time
from FatxFast.tilemap.maze import mazecell, content_filler
from FatxFast.constants import *
import random

class MazeLoadError(Exception): pass

# These are the singleton variables for maze generation
ALGORITHMS = {
"prim": RandomPrim(),
"weave": WeaveBacktrack(),
"backtrack": Backtrack(),
"binary": BinaryTree()
}
WEAVE_OFFSET = 1

class Mazefier(object):

    """Automated plugin for tilemap_editor."""
    def __init__(self, worldeditor, **kwargs):
        self.we = worldeditor
        self.cellcontainer = []
        self.timeseed = None
        self.content_gen = content_filler.MazeContentGen(self)
        
        self.tilewidth = TILEWIDTH
        self.tileheight = TILEHEIGHT

        # TODO: allowing higher and wider walls than 1 tiles
        # Currently adjustable values: floor widht&height 
        self.west_wall_size = [1, 1]
        self.north_wall_size = [1, 1]
        self.floor_width = 1
        self.floor_height = 1
        # Expects this to be in rsc/
        self.tilesheet_imagename = kwargs.get("tileset", "tileset.png")

        #Enumerated tiles in which order they are assumed to be in
        self.tiles = {
            "wall":2, 
            "floor": 1, 
            "start": 3, 
            "end": 4, 
            "warp": 5,
            "floor2": 6,
            "coffee": 7, 
            "collide": 8,
            "warp"+str(WEST): 10,
            "warp"+str(EAST): 9,
            "warp"+str(NORTH): 12,
            "warp"+str(SOUTH): 11}

    def create_object(self, x, y, width, height, type, **kwargs):
        self.we.change_current_layer(OBJECT_LAYER)
        self.we.create_object(x, y, width, height, type, 
                **kwargs)

    def create_cells(self, columns, rows):
        self.cellcontainer = []
        for y in xrange(rows):
            xcells = []
            cell_y_pos = (y*(self.north_wall_size[1]+self.west_wall_size[1]))+1
            for x in xrange(columns):
                cell_x_pos = (
                    x*(self.north_wall_size[0]+self.west_wall_size[0]))+1
                self._create_wall(cell_x_pos-self.west_wall_size[0], 
                    cell_y_pos-self.north_wall_size[1])
                self._create_cell_wall(cell_x_pos, cell_y_pos, NORTH)   
                self._create_cell_wall(cell_x_pos, cell_y_pos, WEST)   
                cell = mazecell.RectMazeCell(cell_x_pos, cell_y_pos, x, y, 
                    width=(self.floor_width+self.west_wall_size[0])*TILEWIDTH, 
                    height=(
                        self.floor_width+self.north_wall_size[1])*TILEHEIGHT)
                if x == columns-1:
                    self._create_wall(cell_x_pos+self.floor_width, 
                        cell_y_pos-self.north_wall_size[1])
                    self._create_cell_wall(cell_x_pos, cell_y_pos, EAST)   
                if y == rows-1:
                    if x == columns-1:
                        self._create_wall(cell_x_pos+self.floor_width, 
                            cell_y_pos+self.floor_height)
                    self._create_wall(cell_x_pos-self.west_wall_size[0], 
                        cell_y_pos+self.west_wall_size[1])
                    self._create_cell_wall(cell_x_pos, cell_y_pos, SOUTH)   
                xcells.append(cell)

            self.cellcontainer.append(xcells)

    def _create_cell_floor(self, x, y, tile):
        self.we.change_current_layer(GROUND_LAYER)
        self.we.change_tiles_in_a_region(tile,
            *self._get_floor_vertices(x, y), entity_id=FLOOR_ID)

    def create_cell_floor_object(self, cell, tile, type, entity_id):
        x, y = cell.x, cell.y
        self._create_cell_floor(x, y, tile)
        self.create_object(cell.rect.x, cell.rect.y, 
            self.floor_width*TILEWIDTH, self.floor_height*TILEHEIGHT, 
            type,  entity_id=entity_id)

    def _destroy_wall_and_collision(self, x, y, direction, new_tile): 
        self.we.change_current_layer(GROUND_LAYER)
        self.we.change_tiles_in_a_region(new_tile, 
            *self._get_wall_vertices(x, y, direction))
        self.we.change_current_layer(COLLIDE_LAYER)
        self.we.change_tiles_in_a_region(None, 
            *self._get_wall_vertices(x, y, direction))

    def _create_wall(self, *vertices):
        self.we.change_current_layer(GROUND_LAYER)
        self.we.change_tiles_in_a_region(self.tiles["wall"], 
            *vertices, entity_id=WALL_ID)
        self.we.change_current_layer(COLLIDE_LAYER)
        self.we.change_tiles_in_a_region(self.tiles["collide"], 
            *vertices, entity_id=WALL_ID)

    def _create_cell_wall(self, x, y, direction):
        self._create_wall(*self._get_wall_vertices(x, y, direction))

    def create_weave(self, cell, direction):
        # TODO: Bug fix for weave object creation with dynamic cells
        x, y, ex, ey = self._get_wall_vertices(cell.x, cell.y, direction)
        self.we.change_current_layer(GROUND_LAYER)
        """self.we.change_tiles_in_a_region(self.tiles["wall"], x, y, ex, ey ,
            entity_id=WARP_ID)"""
        self.we.change_tiles_in_a_region(self.tiles["warp"+str(direction)], 
            x, y, ex, ey , entity_id=WARP_ID) 
        self.we.change_current_layer(OBJECT_LAYER)
        width, height = abs(ex-x), abs(ey-y)
        if not width:
            width = self.west_wall_size[0]
        else:
            width += self.west_wall_size[0]

        if not height:
            height = self.north_wall_size[1]
        else:
            height += self.north_wall_size[1]
        x, y = x*self.tilewidth, y*self.tileheight
        if direction == NORTH:
            y += WEAVE_OFFSET
            ey += WEAVE_OFFSET
        elif direction == SOUTH:
            y -= WEAVE_OFFSET
            ey -= WEAVE_OFFSET
        elif direction == WEST:
            x += WEAVE_OFFSET
            ex += WEAVE_OFFSET
        elif direction == EAST:
            x -= WEAVE_OFFSET
            ex -= WEAVE_OFFSET
        width, height = width*self.tilewidth, height*self.tileheight 
        tile_x, tile_y = self._get_weave_dest(cell, direction)
        if not self.floor_width%2:
            tile_x = tile_x*self.tilewidth
        else:
            tile_x = tile_x*self.tilewidth+(self.tilewidth/2)
        if not self.floor_height%2:
            tile_y = tile_y*self.tileheight
        else:
            tile_y = tile_y*self.tileheight+(self.tileheight/2)
        self.we.create_object(x, y, width, height, "trigger",
            entity_id=WARP_ID, dest_x=tile_x, dest_y=tile_y)
        
    
    def destroy_wall(self, x, y, direction):
        self._destroy_wall_and_collision(x, y, direction, self.tiles["floor2"])

    @property
    def floor_width(self):
        return self._floor_width
    
    @floor_width.setter
    def floor_width(self, value):
        self.north_wall_size[0] = value
        self._floor_width = value

    @property
    def floor_height(self):
        return self._floor_height

    @floor_height.setter
    def floor_height(self, value):
        self.west_wall_size[1] = value
        self._floor_height = value

    def _get_floor_vertices(self, x, y):
        return x, y, x+self.floor_width-1, y+self.floor_height-1

    def get_cell(self, pos):
        """Fetches a cell at a desired location.
        Note that fetching the last element of any dimension is forbidden.
        Reasoning: maze generation algorithms will most likely 
        try to fetch a negative position - [-1] - at some point."""
        if min(pos) > -1:
            try:
                return self.cellcontainer[pos[1]][pos[0]]
            except IndexError:
                pass

    def get_wall(self, cell, direction):
        x, y = self._get_wall_vertices(cell.x, cell.y, direction)[0:2]
        return self._get_wall(x, y)
    
    def _get_wall(self, x, y):
        return self.we.get_tile(x, y, GROUND_LAYER).entity_id == WALL_ID 
    
    def _get_wall_vertices(self, x, y, direction):
        if direction == NORTH:
            pos = (x, y-self.north_wall_size[1], x+self.north_wall_size[0]-1,
                y-self.north_wall_size[1])
        elif direction == WEST:
            pos = (x-self.west_wall_size[0], y, x-self.west_wall_size[0], 
                y+self.west_wall_size[1]-1)
        elif direction == EAST:
            pos = (x+self.floor_width, y, x+self.floor_width, 
                y+self.west_wall_size[1]-1)
        elif direction == SOUTH:
            pos = (x, y+self.floor_height, x+self.north_wall_size[0]-1, 
                y+self.floor_height)
        return pos

    def _get_weave_move(self, direction):
        if direction == NORTH:
            x, y = 0, -2*(self.west_wall_size[1]+self.north_wall_size[1])
        elif direction == WEST:
            x, y = -2*(self.west_wall_size[0]+self.north_wall_size[0]), 0
        elif direction == EAST:
            x, y = 2*(self.west_wall_size[0]+self.north_wall_size[0]), 0
        elif direction == SOUTH:
            x, y = 0, 2*(self.west_wall_size[1]+self.north_wall_size[1])
        return x, y

    def _get_weave_dest(self, cell, direction):
        x, y = self._get_weave_move(direction)
        x, y = x+cell.x, y+cell.y
        return x, y

    def run(self, columns, rows, **kwargs):
        if min(columns,rows) < 1:
            raise MazeLoadError("Illegal maze rows/columns")
        
        self.columns = columns
        self.rows = rows
        self.cellcount = columns*rows
       

        self.timeseed = kwargs.get("timeseed", time.time())
        random.seed(self.timeseed)
        
        start = time.time()
        self.algorithm = ALGORITHMS[kwargs.get("algorithm", 
            random.choice(ALGORITHMS.keys()))] 
        self.algorithm.run(self, **kwargs) 
        self.content_gen.set_start_and_end(
            kwargs.get("start"), kwargs.get("end"))
        items = kwargs.get("items", int(columns*rows/120))
        item_types = kwargs.get("item_types")
        self.content_gen.generate_items(items, item_types)
        self.we.tile_map.properties["timeseed"] = self.timeseed
